import aiohttp
import pytest

from ezmm.request import is_maybe_image_url


@pytest.mark.parametrize("url,expected", [
    ("https://media.cnn.com/api/v1/images/stellar/prod/ap22087057359494.jpg?c=16x9&q=h_653,w_1160,c_fill/f_webp", True),
    ("https://edition.cnn.com/2024/10/30/asia/north-korea-icbm-test-intl-hnk/index.html", False),
    ("https://img.zeit.de/politik/ausland/2024-10/georgien-wahl-stimmauszaehlung-regierungspartei-bild/wide__1000x562__desktop__scale_2", True),
    ("https://upload.wikimedia.org/wikipedia/commons/8/8d/President_Barack_Obama.jpg", True),
    ("https://de.wikipedia.org/wiki/Datei:President_Barack_Obama.jpg", False),  # this is the image's article view
    ("https://bingekulture.com/wp-content/uploads/2021/08/cropped-cropped-logo.fw-removebg-preview.png?w=48", False),  # this URL redirects to a webpage
    ("https://www.popularmechanics.com/_assets/design-tokens/fre/static/icons/play.db7c035.svg?primary=%2523ffffff%20%22Play%22", False),  # this is a vector graphic
    ("https://pixum-cms.imgix.net/7wL8j3wldZEONCSZB9Up6B/d033b7b6280687ce2e4dfe2d4147ff93/fab_mix_kv_perspektive_foto_liegend_desktop__3_.png?auto=compress,format&trim=false&w=2000", True),
    ("https://cdn.pixabay.com/photo/2017/11/08/22/28/camera-2931883_1280.jpg", True),  # image is presented as a binary download stream
    ("https://arxiv.org/pdf/2412.10510", False),  # this is a PDF download stream
    ("https://platform.vox.com/wp-content/uploads/sites/2/2025/04/jack-black-wink-minecraft.avif?quality=90&strip=all&crop=12.5%2C0%2C75%2C100&w=2400", True),
    ("https://media.cnn.com/api/v1/images/stellar/prod/02-overview-of-kursk-training-area-15april2025-wv2.jpg?q=w_1110,c_fill", True)
])
@pytest.mark.asyncio
async def test_is_image_url(url, expected):
    async with aiohttp.ClientSession() as session:
        assert await is_maybe_image_url(url, session) == expected



from ezmm import Image
from ezmm.common.registry import item_registry


def test_registry():
    img = Image("in/roses.jpg")  # Load the image to automatically register it in the registry
    assert item_registry.get(img.reference) is img


def test_cache_miss():
    img1 = Image("in/roses.jpg")

    # Reset cache (as if the registry was just restarted with an existing DB)
    item_registry.cache = dict()

    img2 = Image("in/roses.jpg")

    assert img1 is not img2  # Due to cache miss, but...
    assert img1.id == img2.id



import logging
from io import BytesIO
from pathlib import Path
from typing import Optional

import aiohttp
import pillow_avif  # Keep this import for AVIF support
from PIL.Image import Image as PillowImage, open as pillow_open, new as pillow_new, Resampling

from ezmm.common.items.item import Item
from ezmm.request import request_static
from ezmm.util import to_base64

logger = logging.getLogger("ezMM")
logger.debug(f"`pillow_avif` v{pillow_avif.__version__} loaded for AVIF image support.")


class Image(Item):
    kind = "image"
    _image: Optional[PillowImage] = None

    def __init__(self, file_path: str | Path | None = None,
                 pillow_image: PillowImage | None = None,
                 binary_data: bytes | None = None,
                 source_url: str | None = None,
                 reference: str | None = None):
        assert file_path or pillow_image or binary_data or reference

        if hasattr(self, "id"):
            # The image is already initialized (existing instance returned via __new__())
            return

        if binary_data is not None:
            pillow_image = pillow_open(BytesIO(binary_data))

        if pillow_image is not None:
            pillow_image = _ensure_rgb_mode(pillow_image)
            # Save the image in a temporary folder
            file_path = self._temp_file_path(suffix=".jpg")
            file_path.parent.mkdir(parents=True, exist_ok=True)
            pillow_image.save(file_path)
            self._image = pillow_image

        # free memory and file handle
        self.close()

        super().__init__(file_path,
                         source_url=source_url,
                         reference=reference)

    @property
    def image(self) -> PillowImage:
        """Lazy-loads the PIL image of this Image item."""
        if not self._image:
            with pillow_open(self.file_path) as img:
                self._image = _ensure_rgb_mode(img).copy()
        return self._image

    def get_base64_encoded(self) -> str:
        return to_base64(self.image)

    @property
    def width(self) -> int:
        return self.image.width

    @property
    def height(self) -> int:
        return self.image.height

    def _same(self, other):
        return (
                self.image.mode == other.image.mode and
                self.image.size == other.image.size and
                self.image.tobytes() == other.image.tobytes()
        )

    def as_html(self) -> str:
        img = f'<img src="/{self.file_path.as_posix()}" alt="{self.reference}">'
        if self.source_url:
            return f'<a href="{self.source_url}">{img}</a>'
        else:
            return img

    def close(self):
        if self._image:
            self._image.close()
            self._image = None


def _ensure_rgb_mode(pillow_image: PillowImage) -> PillowImage:
    """Turns any kind of image (incl. PNGs) into RGB mode to make it JPEG-saveable."""
    if pillow_image.mode in ["RGBA", "P"]:
        pillow_image = pillow_image.convert('RGBA')
        converted = pillow_new("RGB", pillow_image.size, (255, 255, 255))
        converted.paste(pillow_image, mask=pillow_image.split()[3])  # 3 is the alpha channel
        return converted
    if pillow_image.mode != "RGB":
        return pillow_image.convert('RGB')
    else:
        return pillow_image


async def download_image(
        image_url: str,
        session: aiohttp.ClientSession,
        ignore_small_images: bool = True,
        max_size: tuple[int, int] = (2048, 2048)
) -> Optional[Image]:
    """Download an image from a URL and return it as an Image object."""
    # TODO: Handle very large images like: https://eoimages.gsfc.nasa.gov/images/imagerecords/144000/144225/campfire_oli_2018312_lrg.jpg
    content = await request_static(image_url, session, get_text=False)
    if content:
        pillow_img = pillow_open(BytesIO(content))

        if pillow_img:
            if pillow_img.width > max_size[0] or pillow_img.height > max_size[1]:
                pillow_img.thumbnail(max_size, Resampling.LANCZOS)  # Preserves aspect ratio

            if not ignore_small_images or (pillow_img.width > 256 and pillow_img.height > 256):
                # TODO: Check for duplicates, i.e., reuse an existing image if it already exists in the registry
                image = Image(pillow_image=pillow_img, source_url=image_url)
                image.relocate(move_not_copy=True)  # Ensure the image is in the temp dir + follows simple naming
                return image



import asyncio
from shutil import copyfile

import aiohttp
import pytest
from PIL import Image as PillowImage

from ezmm import Image, MultimodalSequence, download_image


def test_image_equality():
    # Duplicate image file
    copyfile("in/roses.jpg", "in/roses_copy.jpg")

    img1 = Image("in/roses.jpg")
    img2 = Image("in/roses_copy.jpg")
    assert img1 == img2
    assert img1 is not img2


def test_images_in_sequence():
    img1 = Image("in/roses.jpg")
    img2 = Image("in/garden.jpg")
    seq = MultimodalSequence("The images", img1, img2, "show two beautiful roses and a garden.")
    images = seq.images
    assert len(images) == 2
    assert img1 in images
    assert img2 in images
    assert img1 in seq
    assert img2 in seq


def test_binary():
    # Load tulips image with Pillow
    pillow_img = PillowImage.open("in/tulips.jpg")
    img = Image(pillow_image=pillow_img)
    print(img.file_path)


async def download_img(url):
    async with aiohttp.ClientSession() as session:
        return await download_image(url, session)


@pytest.mark.parametrize("url", [
    "https://media.cnn.com/api/v1/images/stellar/prod/02-overview-of-kursk-training-area-15april2025-wv2.jpg?q=w_1110,c_fill",
    "https://img.zeit.de/wissen/2025-06/hitzewelle-barcelona-temperaturrekord/wide__1000x562__desktop__scale_2",
    "https://factly.in/wp-content/uploads/2025/02/Train-fire-in-Prayagraj-Claim.jpg"
])
def test_download_image(url):
    img = asyncio.run(download_img(url))
    print(img)
    assert isinstance(img, Image)



import sqlite3
from pathlib import Path
from typing import Optional

from ezmm.common.items import Item, ITEM_CLASSES, KIND2ITEM
from ezmm.config import temp_dir
from ezmm.util import parse_ref, normalize_path


class ItemRegistry:
    """Keeps track of all the occurring items efficiently.
    Also holds a cache of already loaded items for efficiency."""
    db_path = Path(temp_dir) / "item_registry.db"

    conn: Optional[sqlite3.Connection] = None
    cur: Optional[sqlite3.Cursor] = None
    cache: dict[tuple[str, int], Item] = dict()

    def __init__(self):
        self.connect()

    def connect(self):
        # Initialize folder, DB, and cache
        if not self.db_path.parent.exists():
            self.db_path.parent.mkdir(exist_ok=True, parents=True)
        self.conn = sqlite3.connect(self.db_path, timeout=10, check_same_thread=False)
        self.conn.execute("PRAGMA journal_mode=WAL;")
        self.cur = self.conn.cursor()
        self._init_db()

    def _init_db(self):
        """Initializes a clean, new DB."""
        for item_cls in ITEM_CLASSES:
            kind = item_cls.kind
            stmt = f"""
                CREATE TABLE IF NOT EXISTS  {kind} (
                    id INTEGER PRIMARY KEY,
                    path TEXT NOT NULL UNIQUE,
                    source_url TEXT NOT NULL
                );
            """
            self.cur.execute(stmt)
            stmt = f"""
                CREATE UNIQUE INDEX IF NOT EXISTS {kind}_path_idx ON {kind}(path);
            """
            self.cur.execute(stmt)
        self.conn.commit()

    def get(self, reference: str = None, kind: str = None, identifier: int = None) -> Optional[Item]:
        """Gets the referenced item object by loading it from the cache or,
        if not in the cache, from the disk."""
        if kind is None or identifier is None:
            assert reference
            kind, identifier = parse_ref(reference)

        # Read from cache
        item = self._get_cached(kind, identifier)

        if item is None:
            # Initialize new item object from DB
            item = self._get_item_by_id(kind, identifier)

        return item

    def get_by_path(self, kind: str, path: Path | str) -> Optional[Item]:
        """Returns the item object located at the path ONLY IF it is
        already registered in the registry."""
        identifier = self._get_id_by_path(kind, path)
        if identifier is not None:
            return self.get(kind=kind, identifier=identifier)

    def add_and_assign_id(self, item: Item):
        """Adds a (partially initialized) item to the registry, if not yet registered.
        Assigns a proper ID to the item and adds it to the cache."""
        if not self.contains(item.kind, item.file_path):
            item.id = self._insert_into_registry(item, item.kind)
        else:
            item.id = self._get_id_by_path(item.kind, item.file_path)
        self._add_to_cache(item)

    def get_cached(self, reference: str = None, kind: str = None, file_path: Path | str= None) -> Optional[Item]:
        if reference:
            kind, identifier = parse_ref(reference)
        else:
            assert kind is not None
            identifier = self._get_id_by_path(kind, file_path)
        return self._get_cached(kind, identifier)

    def _get_id_by_path(self, kind: str, item_path: Path) -> Optional[int]:
        item_path = normalize_path(item_path)
        stmt = f"""
            SELECT id
            FROM {kind}
            WHERE path = ?;
        """
        response = self.cur.execute(stmt, (item_path.as_posix(),))
        result = response.fetchone()
        if result is not None:
            return result[0]
        else:
            return None

    def _get_item_by_id(self, kind: str, identifier: int) -> Optional[Item]:
        stmt = f"""
            SELECT path, source_url
            FROM {kind}
            WHERE id = ?;
        """
        response = self.cur.execute(stmt, (identifier,))
        if result := response.fetchone():
            item_cls = KIND2ITEM[kind]
            return item_cls(file_path=Path(result[0]),
                            source_url=result[1])

    def _insert_into_registry(self, item: Item, kind: str) -> int:
        """Adds the new item directly to the database and returns its assigned ID."""
        stmt = f"""
            INSERT INTO {kind}(path, source_url)
            VALUES (?, ?);
        """
        self.cur.execute(stmt, (item.file_path.as_posix(), item.source_url))
        self.conn.commit()

        stmt = """SELECT last_insert_rowid();"""
        response = self.cur.execute(stmt)
        return response.fetchone()[0]

    def update_file_path(self, item: Item):
        """Updates the path for the corresponding item in the registry."""
        stmt = f"""
            UPDATE {item.kind}
            SET path = ?
            WHERE id = ?;
        """
        self.cur.execute(stmt, (item.file_path.as_posix(), item.id))
        self.conn.commit()

    def contains(self, kind: str, item_path: Path | str) -> bool:
        return self._get_id_by_path(kind, item_path) is not None

    def _get_cached(self, kind: str, identifier: int) -> Optional[Item]:
        """Tries to retrieve the specified item from the cache. Returns
        None if it is not in the cache."""
        return self.cache.get((kind, identifier))

    def _add_to_cache(self, item: Item) -> None:
        """Adds the given item to the cache."""
        self.cache[(item.kind, item.id)] = item

    def close(self):
        self.conn.close()
        self.conn = None

    def reset(self):
        if self.conn:
            self.close()
        self.cache = dict()
        self.connect()


item_registry = ItemRegistry()



from shutil import copyfile

import aiohttp
import pytest

from ezmm import MultimodalSequence, Video, download_video


def test_video():
    vid = Video("in/mountains.mp4")
    print(vid)


def test_video_equality():
    # Duplicate image file
    copyfile("in/mountains.mp4", "in/mountains_copy.mp4")

    vid1 = Video("in/mountains.mp4")
    vid2 = Video("in/mountains_copy.mp4")
    assert vid1 == vid2
    assert vid1 is not vid2


def test_videos_in_sequence():
    vid1 = Video("in/mountains.mp4")
    vid2 = Video("in/snow.mp4")
    seq = MultimodalSequence("The videos", vid1, vid2, "show scenes in the Alps.")
    print(seq)
    videos = seq.videos
    assert len(videos) == 2
    assert vid1 in videos
    assert vid2 in videos
    assert vid1 in seq
    assert vid2 in seq


def test_binary():
    with open("in/mountains.mp4", "rb") as f:
        binary_data = f.read()
    vid = Video(binary_data=binary_data)
    print(vid)


def test_base64():
    vid = Video("in/mountains.mp4")
    print(len(vid.get_base64_encoded()))


@pytest.mark.asyncio
@pytest.mark.parametrize("url", [
    "https://upload.wikimedia.org/wikipedia/commons/transcoded/a/a7/How_to_make_video.webm/How_to_make_video.webm.1080p.vp9.webm",
    "https://demo.unified-streaming.com/k8s/features/stable/video/tears-of-steel/tears-of-steel.ism/.m3u8"
    "https://devstreaming-cdn.apple.com/videos/streaming/examples/adv_dv_atmos/main.m3u8",
])
async def test_download_video(url):
    async with aiohttp.ClientSession() as session:
        vid = await download_video(url, session)
        print(vid)
        assert isinstance(vid, Video)



from ezmm.common.items.item import Item, REF, download_item
from ezmm.common.items.image import Image, download_image
from ezmm.common.items.video import Video, download_video
from ezmm.common.items.audio import Audio

ITEM_CLASSES = [Image, Video, Audio]
KIND2ITEM = {item.kind: item for item in ITEM_CLASSES}
KINDS = [item.kind for item in ITEM_CLASSES]

# Regex patterns
KINDS_ALTERNATIVES = "|".join(KINDS)
ITEM_REF_REGEX = "(" + REF.format(kind=f"(?:{KINDS_ALTERNATIVES})", id="[0-9]+") + ")"  # Captures full ref
ITEM_ID_REGEX = REF.format(kind=f"(?:{KINDS_ALTERNATIVES})", id="([0-9]+)")  # Captures only ID
ITEM_KIND_REGEX = REF.format(kind=f"({KINDS_ALTERNATIVES})", id="[0-9]+")  # Captures only kind
ITEM_KIND_ID_REGEX = REF.format(kind=f"({KINDS_ALTERNATIVES})", id="([0-9]+)")  # Captures kind and ID



from pathlib import Path
from shutil import copyfile

import pytest

from ezmm import Image, Item, download_item


def test_item():
    img = Image("in/roses.jpg")
    print(img)


def test_equality():
    img1 = Image("in/roses.jpg")
    img2 = Image("in/roses.jpg")
    assert img1 == img2


def test_identity():
    img1 = Image("in/roses.jpg")
    img2 = Image("in/roses.jpg")
    assert img1 is img2


def test_inequality():
    img1 = Image("in/roses.jpg")
    img2 = Image("in/garden.jpg")
    assert img1 != img2


def test_reference():
    img1 = Image("in/roses.jpg")
    img2 = Image(reference=img1.reference)
    assert img1 == img2
    assert img1 is img2


def test_from_reference():
    img1 = Image("in/roses.jpg")
    img2 = Item.from_reference(img1.reference)
    assert img1 == img2
    assert img1 is img2


def test_source_default():
    img_path = Path("in/roses.jpg")
    img = Image(img_path)
    assert img_path.absolute().as_uri() == img.source_url


def test_source_custom():
    img_path = Path("in/roses.jpg")
    img = Image(img_path, source_url="https://example.com/image.jpg")
    assert "https://example.com/image.jpg" == img.source_url


def test_relocate_copy():
    img1 = Image("in/roses.jpg")
    img1.relocate()
    new_filepath = img1.file_path.as_posix()
    assert "in/roses.jpg" not in new_filepath
    assert new_filepath.endswith(f"image/{img1.id}.jpg")

    # Loading the original image file should result in an equal but
    # different Image object
    img2 = Image("in/roses.jpg")
    assert img1 == img2
    assert img1 is not img2


def test_relocate_move():
    # Create temp image file from existing
    source_path = Path("in/roses_copy.jpg")
    copyfile("in/roses.jpg", source_path)

    img = Image(source_path)
    img.relocate(move_not_copy=True)
    new_filepath = img.file_path.as_posix()
    assert "in/roses.jpg" not in new_filepath
    assert new_filepath.endswith(f"image/{img.id}.jpg")
    assert not source_path.exists()


@pytest.mark.asyncio
@pytest.mark.parametrize("url", [
        "https://img.zeit.de/wissen/2025-06/hitzewelle-barcelona-temperaturrekord/wide__1000x562__desktop__scale_2",
        "https://upload.wikimedia.org/wikipedia/commons/transcoded/a/a7/How_to_make_video.webm/How_to_make_video.webm.1080p.vp9.webm"
])
async def test_download_items(url):
    item = await download_item(url)
    assert isinstance(item, Item)
    print(item)



import re
import subprocess
import tempfile
from pathlib import Path
from typing import Optional
import imageio_ffmpeg as ffmpeg
from io import BytesIO
import base64
from PIL.Image import Image as PillowImage


def normalize_path(path: Path | str) -> Path:
    """Ensures the path is an (absolute) Path object."""
    return Path(path).absolute()


def get_item_refs(text: str) -> list[str]:
    """Extracts all item references from the text."""
    from ezmm.common.items import ITEM_REF_REGEX
    pattern = re.compile(ITEM_REF_REGEX, re.DOTALL)
    matches = pattern.findall(text)
    return matches


def parse_ref(ref: str) -> (str, int):
    result = parse_item_ref(ref)
    if result is None:
        raise ValueError(f"Invalid item reference: {ref}")
    return result


def parse_item_ref(reference: str) -> Optional[tuple[str, int]]:
    """Returns the first matching kind and identifier from the reference."""
    from ezmm.common.items import ITEM_KIND_ID_REGEX
    pattern = re.compile(ITEM_KIND_ID_REGEX, re.DOTALL)
    result = pattern.findall(reference)
    if len(result) > 0:
        match = result[0]
        return match[0], int(match[1])
    else:
        return None


def is_item_ref(string: str) -> bool:
    """Returns True iff the string represents an item reference."""
    from ezmm.common.items import ITEM_REF_REGEX
    pattern = re.compile(ITEM_REF_REGEX, re.DOTALL)
    return pattern.fullmatch(string) is not None


def validate_references(text: str) -> bool:
    """Verifies that each item reference can be resolved to a registered item."""
    from ezmm.common.registry import item_registry
    refs = get_item_refs(text)
    for ref in refs:
        if item_registry.get(ref) is None:
            return False
    return True


def to_base64(image: PillowImage) -> str:
    """Converts the given Pillow Image to a base64-encoded string."""
    buffered = BytesIO()
    image.save(buffered, format="JPEG")
    return base64.b64encode(buffered.getvalue()).decode('utf-8')


def ts_to_mp4(ts_bytes: bytes) -> bytes:
    """Converts a TS video (bytes) to MP4 (bytes) using FFmpeg with temporary files."""
    with tempfile.NamedTemporaryFile(suffix='.ts', delete=False) as temp_ts_file:
        temp_ts_file.write(ts_bytes)
        temp_ts_path = Path(temp_ts_file.name)

    temp_mp4_path = temp_ts_path.with_suffix('.mp4')

    try:
        # Run FFmpeg to convert TS to MP4
        cmd = [
            ffmpeg.get_ffmpeg_exe(),  # ensure ffmpeg is available in the environment
            "-i", str(temp_ts_path),
            "-c:v", "copy",
            "-c:a", "copy",
            "-f", "mp4",
            str(temp_mp4_path)
        ]
        result = subprocess.run(cmd, stderr=subprocess.PIPE, stdout=subprocess.PIPE)

        if result.returncode != 0:
            raise RuntimeError(f"FFmpeg error:\n{result.stderr.decode(errors='ignore')}")

        # Read and return the MP4 bytes
        with open(temp_mp4_path, "rb") as mp4_file:
            mp4_bytes = mp4_file.read()

    finally:
        # Clean up temporary files
        temp_ts_path.unlink(missing_ok=True)
        temp_mp4_path.unlink(missing_ok=True)

    return mp4_bytes



import base64
import logging
from pathlib import Path
from typing import Optional
from urllib.parse import urljoin

import aiohttp
import cv2
import m3u8
import numpy as np

from ezmm.common.items.item import Item
from ezmm.request import fetch_headers, HEADERS
from ezmm.util import ts_to_mp4, to_base64

logger = logging.getLogger("ezMM")


class Video(Item):
    kind = "video"
    _video: Optional[cv2.VideoCapture] = None

    def __init__(self, file_path: str | Path = None,
                 binary_data: bytes = None,
                 source_url: str = None,
                 reference: str = None):
        assert file_path or binary_data or reference

        if hasattr(self, "id"):
            return

        if binary_data:
            # Save binary data to temporary file
            file_path = self._temp_file_path(suffix=".mp4")
            file_path.parent.mkdir(parents=True, exist_ok=True)
            with open(file_path, 'wb') as f:
                f.write(binary_data)

        super().__init__(file_path,
                         source_url=source_url,
                         reference=reference)

    @property
    def video(self) -> cv2.VideoCapture:
        """Lazy-loads the video capture of this Video item."""
        if not self._video:
            self._video = cv2.VideoCapture(str(self.file_path))
        return self._video

    @property
    def width(self) -> int:
        return int(self.video.get(cv2.CAP_PROP_FRAME_WIDTH))

    @property
    def height(self) -> int:
        return int(self.video.get(cv2.CAP_PROP_FRAME_HEIGHT))

    @property
    def frame_count(self) -> int:
        return int(self.video.get(cv2.CAP_PROP_FRAME_COUNT))

    @property
    def fps(self) -> float:
        return self.video.get(cv2.CAP_PROP_FPS)

    @property
    def duration(self) -> float:
        """Returns the duration of the video in seconds."""
        return self.frame_count / self.fps

    @property
    def bytes(self) -> bytes:
        """Returns the video as bytes."""
        return self.file_path.read_bytes()

    def _get_frames(self) -> list[bytes]:
        """Returns the list of all frames from the video."""
        frames = []
        while self.video.isOpened():
            success, frame = self.video.read()
            if not success:
                break
            _, frame = cv2.imencode(".jpeg", frame)
            frames.append(frame)
        self.video.release()
        return frames

    def sample_frames(self, n_frames: int = 5) -> list[bytes]:
        """Returns the specified number of frames sampled evenly spaced from the video.
        Always includes the first frame. Includes the last frame if n_frames > 1."""
        assert n_frames > 0, "Number of frames must be greater than 0."
        frames = self._get_frames()
        frame_ids = np.linspace(0, len(frames) - 1, n_frames, dtype=int)
        sampled = [frames[i] for i in frame_ids]
        return sampled

    def get_base64_encoded(self, n_frames: int = 5) -> list[str]:
        """Returns base64-encoded frames, evenly sampled from the video."""
        frames = self.sample_frames(n_frames)
        frames_encoded = [base64.b64encode(frame).decode("utf-8") for frame in frames]
        return frames_encoded

    def _same(self, other):
        return (
                self.width == other.width and
                self.height == other.height and
                self.frame_count == other.frame_count and
                self.file_path.read_bytes() == other.file_path.read_bytes()
        )

    def as_html(self) -> str:
        return f'<video controls src="/{self.file_path.as_posix()}"></video>'

    def close(self):
        if self._video:
            self._video.release()
            self._video = None


async def download_video(
        video_url: str,
        session: aiohttp.ClientSession
) -> Optional[Video]:
    """Downloads the linked video (stream) and returns it as a Video object."""

    try:
        headers = await fetch_headers(video_url, session, timeout=3)
        content_type = headers.get('Content-Type') or headers.get('content-type')
        if content_type.startswith("video/"):
            return await download_video_file(video_url, session)
        elif content_type == "application/vnd.apple.mpegurl":
            return await download_hls_video(video_url, session)
        else:
            logger.warning(f"Cannot download video! Unable to handle content type: {content_type}.")

    except Exception as e:
        logger.debug(f"Error downloading video from {video_url}"
                     f"\n{type(e).__name__}: {e}")


async def download_video_file(
        video_url: str,
        session: aiohttp.ClientSession
) -> Optional[Video]:
    """Download a single video file from a URL and return it as a Video object."""
    try:
        async with session.get(video_url, headers=HEADERS) as response:
            if response.status == 200:
                content = await response.read()
                video = Video(binary_data=content, source_url=video_url)
                video.relocate(move_not_copy=True)
                return video
    except Exception as e:
        logger.debug(f"Error downloading video file from {video_url}"
                     f"\n{type(e).__name__}: {e}")


async def download_hls_video(
        playlist_url: str,
        session: aiohttp.ClientSession
) -> Optional[Video]:
    """Download an HTTP Live Streaming (HLS) video from a playlist URL and return it as a Video object."""
    try:
        # Download the m3u8 playlist file
        async with session.get(playlist_url, headers=HEADERS) as response:
            if response.status != 200:
                logger.debug(f"Failed to download playlist: {response.status}")
                return None
            playlist_content = await response.text()

        playlist = m3u8.loads(playlist_content)
        base_url = playlist_url.rsplit('/', 1)[0] + '/'

        # Check if this is a master playlist (contains variant playlists)
        if playlist.is_variant:
            # Choose the highest quality variant
            best_playlist = playlist.playlists[-1]  # Usually the last one is of highest quality

            # Manually construct the absolute URL for the variant playlist
            variant_url = urljoin(base_url, best_playlist.uri)

            # Download the variant playlist
            async with session.get(variant_url, headers=HEADERS) as var_response:
                if var_response.status != 200:
                    logger.error(f"Failed to download variant playlist: {var_response.status}")
                    return None
                variant_content = await var_response.text()

            # Parse the variant playlist
            variant_playlist = m3u8.loads(variant_content)
            playlist = variant_playlist  # Use this for segment downloads

            # Update base_url for segment downloads
            base_url = variant_url.rsplit('/', 1)[0] + '/'

        # Download all segments
        video_segments = []

        for i, segment in enumerate(playlist.segments):
            # Construct full URL for the segment
            if segment.uri.startswith('http'):
                segment_url = segment.uri
            else:
                segment_url = urljoin(base_url, segment.uri)

            # Download the segment with SSL disabled
            try:
                async with session.get(segment_url, headers=HEADERS) as seg_response:
                    if seg_response.status == 200:
                        segment_data = await seg_response.read()
                        video_segments.append(segment_data)
            except Exception as e:
                logger.debug(f"Failed to download segment {i} from {segment_url}: {e}")

        # Combine all segments
        if video_segments:
            ts_bytes = b''.join(video_segments)
            mp4_bytes = ts_to_mp4(ts_bytes)

            # Create Video object with MP4 content
            video = Video(binary_data=mp4_bytes, source_url=playlist_url)
            video.relocate(move_not_copy=True)
            return video

    except Exception as e:
        logger.debug(f"Error downloading HLS video from {playlist_url}"
                     f"\n{type(e).__name__}: {e}")

    return None


async def download_vid(url):
    async with aiohttp.ClientSession() as session:
        return await download_video(url, session)


if __name__ == "__main__":
    import asyncio

    video = asyncio.run(
        download_vid("https://upload.wikimedia.org/wikipedia/commons/transcoded/a/a7/How_to_make_video.webm/How_to_make_video.webm.1080p.vp9.webm"))
    print(video)



import sys

from ezmm.common.items import *
from ezmm.common.multimodal_sequence import MultimodalSequence
from ezmm.common.registry import ItemRegistry
import logging

APP_NAME = "ezMM"

# Set up logger
logger = logging.getLogger(APP_NAME)
logger.setLevel(logging.DEBUG)

# Only add handler if none exists (avoid duplicate logs on rerun)
if not logger.hasHandlers():
    handler = logging.StreamHandler(sys.stdout)
    formatter = logging.Formatter('[%(levelname)s]: %(message)s')
    handler.setFormatter(formatter)
    logger.addHandler(handler)



from __future__ import annotations

from typing import Sequence
from markdown import markdown

from ezmm.common.items import Image, Audio, Video
from ezmm.common.items.item import Item, resolve_references_from_sequence
from ezmm.ui.common import SEQ_PATH


class MultimodalSequence:
    """A sequence of data of any kind. Can be serialized to a string where each
    non-verbalizable element is referenced in-place, e.g., `<image:0>` for image
    with ID 0."""
    data: list[str | Item]

    def __init__(self, *args: str | Item | MultimodalSequence | Sequence[str | Item | None] | None):
        data = args[0] if len(args) == 1 else list(args)
        if isinstance(data, (str, Item)):
            data = [data]
        elif isinstance(data, MultimodalSequence):
            data = data.data
        self.data = resolve_references_from_sequence(data) if data else []

    @property
    def images(self) -> list[Image]:
        return [item for item in self.data if isinstance(item, Image)]

    @property
    def videos(self) -> list[Video]:
        return [item for item in self.data if isinstance(item, Video)]

    @property
    def audios(self) -> list[Audio]:
        return [item for item in self.data if isinstance(item, Audio)]

    def has_images(self) -> bool:
        return len(self.images) > 0

    def has_videos(self) -> bool:
        return len(self.videos) > 0

    def has_audios(self) -> bool:
        return len(self.audios) > 0

    def to_list(self):
        return self.data

    def as_html(self) -> str:
        """Returns the sequence as HTML code."""
        htmls = []
        for item in self:
            if isinstance(item, Item):
                html = f'<div class="media-container">{item.as_html()}</div>'
                htmls.append(html)
            else:
                htmls.append(markdown(item))
        return " ".join(htmls)

    def unique_items(self) -> set[Item]:
        """Returns the set of all items (not strings) occurring in the sequence."""
        return set([item for item in self if isinstance(item, Item)])

    def __str__(self):
        """Turns itself into a single string where each item is replaced by its reference."""
        substrings = []
        for item in self:
            if isinstance(item, Item):
                substrings.append(item.reference)
            else:
                substrings.append(item)
        return " ".join(substrings)

    def __repr__(self):
        return f"MultimodalSequence(str_len={len(self.__str__())}, n_items={len(self.data)})"

    def __iter__(self):
        return iter(self.data)

    def __len__(self):
        return len(self.data)

    def __getitem__(self, index):
        return self.data[index]

    def __eq__(self, other):
        return isinstance(other, MultimodalSequence) and self.data == other.data

    def __hash__(self):
        return hash(str(self))

    def __bool__(self):
        return len(self) > 0

    def render(self):
        """Saves the given MultimodalSequence to a .md file in SEQ_PATH to make it
        viewable through a link in the browser. Starts a server (if not started yet)
        that serves the UI. If `blocking` is False, the server is started in a separate
        process."""
        # Create sequences directory if it doesn't exist
        SEQ_PATH.mkdir(parents=True, exist_ok=True)

        # Generate 8-char ID string for the sequence
        seq_id = str(abs(hash(self)))[:8]

        # Save to file with unique name
        file_path = SEQ_PATH / f"{seq_id}.md"
        file_path.write_text(str(self), encoding="utf-8")

        # Start the server
        from ezmm.ui.main import run_server
        print(f"You can view the sequence at http://localhost:7878/sequence/{seq_id}")
        run_server()



import asyncio
import re
from abc import ABC
from datetime import datetime
from pathlib import Path
from shutil import copyfile, move
from typing import Sequence, Optional

import aiohttp

from ezmm.config import temp_dir
from ezmm.request import is_maybe_image_url, is_maybe_video_url
from ezmm.util import is_item_ref, normalize_path

REF = "<{kind}:{id}>"  # General reference template, defining the reference syntax


class Item(ABC):
    """An element of MultimodalSequences. The data of each item is saved in an individual file."""
    kind: str  # Specifies the type of the item (image, video, ...)
    id: int  # Unique identifier of this item within its kind
    file_path: Path  # The path (relative to workdir) to the file where the data of this item is stored
    source_url: str  # The (web or file) URL pointing at the Item data's origin

    def __new__(cls, file_path: Path | str = None, source_url: str = None, reference: str = None, **kwargs):
        """Checks if there already exists an instance of the item with the given reference.
        If yes, returns the existing reference. Otherwise, instantiates a new one."""

        if file_path or reference:
            # Look up an existing instance instead of creating a new one
            from ezmm.common.registry import item_registry
            item = item_registry.get_cached(reference=reference, kind=cls.kind, file_path=file_path)
            if item:
                item.source_url = source_url or item.source_url
                return item
            elif reference:
                raise ValueError(f"No item with reference '{reference}'.")

        return super().__new__(cls)

    def __init__(self, file_path: Path | str, source_url: str = None, reference: str = None):
        if hasattr(self, "id"):
            # The item is already initialized (existing instance returned via __new__())
            return
        self.file_path = normalize_path(file_path)
        self.source_url = source_url or self.file_path.absolute().as_uri()
        from ezmm.common.registry import item_registry
        item_registry.add_and_assign_id(self)  # Ensure the item is registered and get an ID assigned

    @property
    def reference(self) -> str:
        return REF.format(kind=self.kind, id=self.id)

    def _same(self, other) -> bool:
        """Compares the content data with the other item for equality."""
        raise NotImplementedError

    @staticmethod
    def from_reference(reference: str) -> Optional["Item"]:
        from ezmm.common.registry import item_registry
        return item_registry.get(reference)

    def close(self):
        """Closes any resources held by this item."""
        pass

    def as_html(self) -> str:
        """Returns the item as HTML code."""
        return f"<p>Item {self.reference} does not support HTML yet.</p>"

    def relocate(self, move_not_copy=False):
        """Copies the item's file to the temp_dir if not
        located there already. Moves it instead if move=True."""
        new_path = self._default_file_path(suffix=self.file_path.suffix)
        if self.file_path != new_path:
            # Ensure the target directory exists
            new_path.parent.mkdir(parents=True, exist_ok=True)

            # Move/copy file to target directory
            self.close()
            move(self.file_path, new_path) if move_not_copy else copyfile(self.file_path, new_path)

            # Update the file path to the new location
            self.file_path = new_path
            from ezmm.common.registry import item_registry
            item_registry.update_file_path(self)

    def _temp_file_path(self, suffix: str = "") -> Path:
        """Used when the item's ID is not set yet."""
        filename = datetime.now().strftime("%Y-%m-%d_%H-%M-%S-%f") + suffix
        return normalize_path(temp_dir / "items" / filename)

    def _default_file_path(self, suffix: str = "") -> Path:
        """Only usable after item initialization."""
        default_filename = str(self.id) + suffix
        return normalize_path(temp_dir / self.kind / default_filename)

    def __eq__(self, other):
        return (self is other or
                isinstance(other, Item) and (
                        self.kind == other.kind and self.id == other.id or  # Should never trigger
                        self._same(other)
                ))

    def __hash__(self):
        # TODO: Make hash content-dependent => identify known items by hash
        return hash((self.kind, self.id))


def resolve_references_from_sequence(seq: Sequence[str | Item]) -> list[str | Item]:
    """Identifies all item references within the sequence and replaces them with
    an instance of the referenced item. Returns the (interleaved) list of
    strings and items."""
    processed = []
    for item in seq:
        if isinstance(item, str):
            if item.strip(" "):  # Drop excess whitespaces
                resolved = resolve_references_from_string(item)
                processed.extend(resolved)
        elif item:  # Drop Nones
            processed.append(item)
    return processed


def resolve_references_from_string(string: str) -> list[str | Item]:
    """Identifies all item references within the string and replaces them with
    an instance of the referenced item. Returns the (interleaved) list of
    strings and items."""
    from ezmm.common.registry import item_registry
    from ezmm.common.items import ITEM_REF_REGEX
    ref_regex = rf"\s?{ITEM_REF_REGEX}\s?"  # Extend to optional whitespaces before and after the ref
    split = re.split(ref_regex, string)
    # Replace each reference with its actual item object
    for i in range(len(split)):
        substr = split[i]
        if is_item_ref(substr):
            item = item_registry.get(substr)
            if item is not None:
                split[i] = item
    return split


async def download_item(url: str,
                        session: Optional[aiohttp.ClientSession] = None,
                        ignore_small_images: bool = True) -> Optional[Item]:
    """Downloads the item from the given URL and returns an instance of the
    corresponding item class. Reuses a session if provided."""

    own_session = session is None
    if own_session:
        session = aiohttp.ClientSession()

    try:
        if await is_maybe_image_url(url, session):
            from ezmm.common.items.image import download_image
            return await download_image(url, ignore_small_images=ignore_small_images, session=session)
        if await is_maybe_video_url(url, session):
            from ezmm.common.items.video import download_video
            return await download_video(url, session)
        # TODO: Handle audios
    finally:
        if own_session:
            await session.close()

if __name__ == "__main__":
    result = asyncio.run(download_item("https://media.cnn.com/api/v1/images/stellar/prod/02-overview-of-kursk-training-area-15april2025-wv2.jpg?q=w_1110,c_fill"))
    print(result)



from ezmm import MultimodalSequence, Image, Video


def test_multimodal_sequence():
    seq = MultimodalSequence("This is just some text.")
    print(seq)

    img = Image("in/roses.jpg")
    seq = MultimodalSequence("The image", img, "shows two beautiful roses.")
    print(seq)


def test_seq_equality():
    img = Image("in/roses.jpg")
    seq1 = MultimodalSequence("The image", img, "shows two beautiful roses.")
    seq2 = MultimodalSequence(["The image", img, "shows two beautiful roses."])
    seq3 = MultimodalSequence(f"The image {img.reference} shows two beautiful roses.")
    assert seq1 == seq2
    assert seq1 == seq3
    assert seq1 is not seq2
    assert seq1 is not seq3
    assert seq2 is not seq3


def test_seq_inequality():
    img1 = Image("in/roses.jpg")
    img2 = Image("in/garden.jpg")
    seq1 = MultimodalSequence("The image", img1, "is nice.")
    seq2 = MultimodalSequence("The image", img2, "is nice.")
    assert seq1 != seq2


def test_list_comprehension():
    img = Image("in/roses.jpg")
    seq = MultimodalSequence("The image", img, "shows two beautiful roses.")
    assert seq[0] == "The image"
    assert seq[1] == img
    assert seq[2] == "shows two beautiful roses."


def test_empty():
    seq0 = MultimodalSequence()
    seq1 = MultimodalSequence("")
    seq2 = MultimodalSequence(" ")
    seq3 = MultimodalSequence([])
    seq4 = MultimodalSequence(None)
    seq5 = MultimodalSequence([None])
    assert seq0 == seq1 == seq2 == seq3 == seq4 == seq5


def test_bool_false():
    seq0 = MultimodalSequence()
    seq1 = MultimodalSequence(None)
    seq2 = MultimodalSequence([])
    seq3 = MultimodalSequence("")
    seq4 = MultimodalSequence(None, None)
    assert not seq0
    assert not seq1
    assert not seq2
    assert not seq3
    assert not seq4


def test_bool_true():
    seq0 = MultimodalSequence("This is just some text.")
    seq1 = MultimodalSequence("The image", Image("in/roses.jpg"), "shows two beautiful roses.")
    assert seq0
    assert seq1


# def test_render():
#     seq = MultimodalSequence(
#         "The image",
#         Image("in/roses.jpg"),
#         "shows two beautiful roses while the video",
#         Video("in/mountains.mp4"),
#         "shows a nice mountain view."
#     )
#     seq.render()



from typing import Optional

import aiohttp
import asyncio
import certifi
import ssl

HEADERS = {
    "User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) "
                  "AppleWebKit/537.36 (KHTML, like Gecko) "
                  "Chrome/123.0.0.0 Safari/537.36",
}

ssl_context = ssl.create_default_context(cafile=certifi.where())

async def is_maybe_image_url(url: str, session: aiohttp.ClientSession) -> bool:
    """Returns True iff the URL points at an accessible _pixel_ image file
    or if the content type is a binary download stream."""
    try:
        headers = await fetch_headers(url, session, timeout=3)
        content_type = headers.get('Content-Type') or headers.get('content-type')
        if content_type.startswith("image/"):
            # Surely an image
            return (not "svg" in content_type and
                    not "eps" in content_type)
        else:
            # If the content is a binary download stream, it may encode an image
            # but also something else. This is a case of "maybe an image"
            return content_type == "binary/octet-stream"

    except Exception:
        return False


async def is_maybe_video_url(url: str, session: aiohttp.ClientSession) -> bool:
    """Returns True iff the URL points at an accessible video file/stream."""
    try:
        headers = await fetch_headers(url, session, timeout=3)
        content_type = headers.get('Content-Type') or headers.get('content-type')
        if content_type.startswith("video/") or content_type == "application/vnd.apple.mpegurl":
            # Surely a video
            return True
        else:
            # If the content is a binary download stream, it may encode a video
            # but also something else. This is a case of "maybe a video"
            return content_type == "binary/octet-stream"

    except Exception:
        return False


async def fetch_headers(url, session: aiohttp.ClientSession, **kwargs) -> dict:
    async with session.head(url, ssl=ssl_context, headers=HEADERS, **kwargs) as response:
        response.raise_for_status()
        return dict(response.headers)


async def request_static(url: str,
                         session: aiohttp.ClientSession,
                         get_text: bool = True,
                         **kwargs) -> Optional[str | bytes]:
    """Downloads the static page from the given URL using aiohttp. If `get_text` is True,
    returns the HTML as text. Otherwise, returns the raw binary content (e.g. an image)."""
    # TODO: Handle web archive URLs
    if url:
        url = str(url)
        try:
            async with session.get(url, timeout=10, headers=HEADERS, allow_redirects=True,
                                   raise_for_status=True, ssl=ssl_context, **kwargs) as response:
                if get_text:
                    return await response.text()  # HTML string
                else:
                    return await stream(response)  # Binary data
        except asyncio.TimeoutError:
            pass  # Server too slow
        except UnicodeError:
            pass  # Page not readable
        except (aiohttp.ClientOSError, aiohttp.ClientConnectorError):
            pass  # Page not available anymore
        except aiohttp.ClientResponseError as e:
            if e.status in [403, 404, 429, 500, 502, 503]:
                # 403: Forbidden access
                # 404: Not found
                # 429: Too many requests
                # 500: Server error
                # 502: Bad gateway
                # 503: Service unavailable (e.g. rate limit)
                pass
            else:
                print(f"\rFailed to retrieve page.\n\t{type(e).__name__}: {e}")
        except Exception as e:
            print(f"\rFailed to retrieve page at {url}.\n\tReason: {type(e).__name__}: {e}")


async def stream(response: aiohttp.ClientResponse, chunk_size: int = 1024) -> bytes:
    data = bytearray()
    async for chunk in response.content.iter_chunked(chunk_size):
        data.extend(chunk)
    return bytes(data)  # Convert to immutable bytes if needed



import socket
import uvicorn
from fastapi import FastAPI
from fastapi.responses import HTMLResponse
from fastapi.staticfiles import StaticFiles
from fastapi.templating import Jinja2Templates
from starlette.requests import Request

from ezmm import MultimodalSequence
from ezmm.ui.common import SEQ_PATH, TEMP_PATH, PROJECT_ROOT

app = FastAPI()

# Mount templates files (CSS, JS, etc. if needed)
app.mount("/templates", StaticFiles(directory=PROJECT_ROOT / "ezmm/ui/templates"), name="templates")
app.mount("/static", StaticFiles(directory=PROJECT_ROOT / "ezmm/ui/static"), name="static")
app.mount("/temp", StaticFiles(directory=TEMP_PATH), name="temp")
app.mount("/in", StaticFiles(directory=PROJECT_ROOT / "in"), name="in")

# Jinja2 templates directory
templates = Jinja2Templates(directory=PROJECT_ROOT / "ezmm/ui/templates")


@app.get("/sequence/{seq_id}", response_class=HTMLResponse)
async def show_claim(request: Request, seq_id: int):
    """Reads the specified MultimodalSequence saved in static/sequences
    and displays it."""
    file_path = SEQ_PATH / f"{seq_id}.md"
    seq_str = file_path.read_text(encoding="utf-8")
    sequence = MultimodalSequence(seq_str)

    return templates.TemplateResponse("sequence.html", {
        "request": request,
        "sequence": sequence,
        "seq_id": seq_id,
    })


@app.get("/", response_class=HTMLResponse)
async def overview(request: Request):
    """Shows a list of all saved MultimodalSequences."""
    sequences = {}
    for seq_file in SEQ_PATH.glob("*.md"):
        seq_id = seq_file.stem
        seq_str = seq_file.read_text(encoding="utf-8")
        seq = MultimodalSequence(seq_str)
        sequences[seq_id] = seq

    return templates.TemplateResponse("overview.html", {
        "request": request,
        "sequences": sequences
    })


def is_port_in_use(port: int) -> bool:
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        try:
            s.bind(('0.0.0.0', port))
            return False
        except socket.error:
            return True


def run_server():
    if not is_port_in_use(7878):
        uvicorn.run("ezmm.ui.main:app", host="0.0.0.0", port=7878, reload=True)


if __name__ == "__main__":
    run_server()
